@@ -25,32 +25,23 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 
 DB-API 2.0 specification: http://www.python.org/dev/peps/pep-0249/
 """
+from __future__ import absolute_import, unicode_literals
 
 import sys
 import time
 import datetime
 import re
-from django.utils import six
 
-try:
-    import decimal
-except ImportError:
-    from django.utils import _decimal as decimal
+import decimal
 
 from django.conf import settings
 from django.db.utils import IntegrityError as DjangoIntegrityError, \
     DatabaseError as DjangoDatabaseError
 
-try:
-    from django.utils import timezone
-except ImportError:
-    # timezone added in Django 1.4
-    timezone = None
+from django.utils import six
+from django.utils import timezone
 
-import pythoncom
-import win32com.client
-
-from ado_consts import *
+from .ado_consts import *
 
 # DB API default values
 apilevel = '2.0'
@@ -84,7 +75,7 @@ class MultiMap(object):
         self.storage = dict()
         self.default = default
 
-        for keys, value in mapping.iteritems():
+        for keys, value in six.iteritems(mapping):
             for key in keys:
                 self.storage[key] = value
 
@@ -94,14 +85,15 @@ class MultiMap(object):
 
 def standardErrorHandler(connection, cursor, errorclass, errorvalue):
     err = (errorclass, errorvalue)
-    connection.messages.append(err)
+    if connection is not None:
+        connection.messages.append(err)
     if cursor is not None:
         cursor.messages.append(err)
     raise errorclass(errorvalue)
 
 
-class Error(StandardError): pass
-class Warning(StandardError): pass
+class Error(Exception if six.PY3 else StandardError): pass
+class Warning(Exception if six.PY3 else StandardError): pass
 
 class InterfaceError(Error): pass
 
@@ -114,10 +106,10 @@ class IntegrityError(DatabaseError, DjangoIntegrityError): pass
 class DataError(DatabaseError): pass
 class NotSupportedError(DatabaseError): pass
 
-class FetchFailedError(Error): 
+class FetchFailedError(Error):
     """
     Error is used by RawStoredProcedureQuerySet to determine when a fetch
-    failed due to a connection being closed or there is no record set 
+    failed due to a connection being closed or there is no record set
     returned.
     """
     pass
@@ -129,6 +121,13 @@ class _DbType(object):
     def __eq__(self, other): return other in self.values
     def __ne__(self, other): return other not in self.values
 
+_re_find_password = re.compile('(pwd|password)=[^;]*;', re.IGNORECASE)
+
+def mask_connection_string_password(s, mask='******'):
+    """
+    Look for a connection string password in 's' and mask it.
+    """
+    return re.sub(_re_find_password, '\g<1>=%s;' % mask, s)
 
 def connect(connection_string, timeout=30, use_transactions=None):
     """Connect to a database.
@@ -137,6 +136,9 @@ def connect(connection_string, timeout=30, use_transactions=None):
         http://www.connectionstrings.com/?carrier=sqlserver2005
     timeout -- A command timeout value, in seconds (default 30 seconds)
     """
+    # Inner imports to make this module importable on non-Windows platforms.
+    import pythoncom
+    import win32com.client
     try:
         pythoncom.CoInitialize()
         c = win32com.client.Dispatch('ADODB.Connection')
@@ -149,7 +151,11 @@ def connect(connection_string, timeout=30, use_transactions=None):
             useTransactions = use_transactions
         return Connection(c, useTransactions)
     except Exception as e:
-        raise OperationalError(e, "Error opening connection: " + connection_string)
+        raise OperationalError(e,
+            "Error opening connection: {0}".format(
+                mask_connection_string_password(connection_string)
+            )
+        )
 
 def _use_transactions(c):
     """Return True if the given ADODB.Connection supports transactions."""
@@ -184,53 +190,54 @@ def format_parameters(parameters, show_value=False):
 
     return '[' + ', '.join(desc) + ']'
 
+def format_decimal_as_string(value):
+    """
+    Convert a decimal.Decimal to a fixed point string. Code borrowed from
+    Python's moneyfmt recipe.
+    https://docs.python.org/2/library/decimal.html#recipes
+    """
+    sign, digits, exp = value.as_tuple()
+    result = []
+    digits = list(map(str, digits))
+    build, next = result.append, digits.pop
+    for i in range(-exp):
+        build(next() if digits else '0')
+    build('.')
+    if not digits:
+        build('0')
+    while digits:
+        build(next())
+    if sign:
+        build('-')
+    return ''.join(reversed(result))
+
 def _configure_parameter(p, value):
     """Configure the given ADO Parameter 'p' with the Python 'value'."""
     if p.Direction not in [adParamInput, adParamInputOutput, adParamUnknown]:
         return
 
-    if isinstance(value, basestring):
+    if isinstance(value, six.string_types):
         p.Value = value
-        #p.Size = len(value)
+        p.Size = len(value)
 
-    elif isinstance(value, buffer):
+    elif isinstance(value, six.memoryview):
         p.Size = len(value)
         p.AppendChunk(value)
 
     elif isinstance(value, decimal.Decimal):
-        p.Value = value
-        exponent = value.as_tuple()[2]
-        digit_count = len(value.as_tuple()[1])
-        
-        if exponent == 0:
-            p.NumericScale = 0
-            p.Precision =  digit_count
-        elif exponent < 0:
-            p.NumericScale = -exponent
-            p.Precision = digit_count
-            if p.Precision < p.NumericScale:
-                p.Precision = p.NumericScale            
-        elif exponent > 0:
-            p.NumericScale = 0
-            p.Precision = digit_count + exponent
+        p.Type = adBSTR
+        p.Value = format_decimal_as_string(value)
 
     elif isinstance(value, datetime.datetime):
-        if getattr(settings, 'USE_TZ', False) and value.tzinfo:
-            value = value.astimezone(timezone.utc)
         p.Type = adBSTR
-        s = value.isoformat(' ')
+        if timezone.is_aware(value):
+            value = timezone.make_naive(value, timezone.utc)
+        # Strip '-' so SQL Server parses as YYYYMMDD for all languages/formats
+        s = value.isoformat(' ' if six.PY3 else b' ').replace('-', '')
         p.Value = s
         p.Size = len(s)
 
     elif isinstance(value, datetime.time):
-        if getattr(settings, 'USE_TZ', False) and value.tzinfo:
-            value = value.astimezone(timezone.utc)
-        p.Type = adBSTR
-        s = value.isoformat()
-        p.Value = s
-        p.Size = len(s)
-
-    elif isinstance(value, datetime.date):
         p.Type = adBSTR
         s = value.isoformat()
         p.Value = s
@@ -251,19 +258,20 @@ class Connection(object):
         self.messages = []
         self.adoConn.CursorLocation = defaultCursorLocation
         self.supportsTransactions = useTransactions
+        self.transaction_level = 0 # 0 == Not in a transaction, at the top level
 
         if self.supportsTransactions:
             self.adoConn.IsolationLevel = defaultIsolationLevel
-            self.adoConn.BeginTrans() # Disables autocommit per DBPAI
+            self.transaction_level = self.adoConn.BeginTrans() # Disables autocommit per DBPAI
 
     def set_autocommit(self, value):
         if self.supportsTransactions == (not value):
             return
         if self.supportsTransactions:
-            self.adoConn.RollbackTrans() # Disables autocommit per DBPAI
+            self.transaction_level = self.adoConn.RollbackTrans() # Disables autocommit per DBPAI
         else:
             self.adoConn.IsolationLevel = defaultIsolationLevel
-            self.adoConn.BeginTrans() # Disables autocommit per DBPAI
+            self.transaction_level = self.adoConn.BeginTrans() # Disables autocommit per DBPAI
         self.supportsTransactions = not value
 
     def _raiseConnectionError(self, errorclass, errorvalue):
@@ -275,7 +283,7 @@ class Connection(object):
     def _close_connection(self):
         """Close the underlying ADO Connection object, rolling back an active transaction if supported."""
         if self.supportsTransactions:
-            self.adoConn.RollbackTrans()
+            self.transaction_level = self.adoConn.RollbackTrans()
         self.adoConn.Close()
 
     def close(self):
@@ -286,6 +294,8 @@ class Connection(object):
         except Exception as e:
             self._raiseConnectionError(InternalError, e)
         self.adoConn = None
+         # Inner import to make this module importable on non-Windows platforms.
+        import pythoncom
         pythoncom.CoUninitialize()
 
     def commit(self):
@@ -299,7 +309,7 @@ class Connection(object):
             return
 
         try:
-            self.adoConn.CommitTrans()
+            self.transaction_level = self.adoConn.CommitTrans()
             if not(self.adoConn.Attributes & adXactCommitRetaining):
                 #If attributes has adXactCommitRetaining it performs retaining commits that is,
                 #calling CommitTrans automatically starts a new transaction. Not all providers support this.
@@ -311,20 +321,17 @@ class Connection(object):
     def rollback(self):
         """Abort a pending transaction."""
         self.messages = []
-        #if not self.supportsTransactions:
-        #    self._raiseConnectionError(NotSupportedError, None)
         with self.cursor() as cursor:
             cursor.execute("select @@TRANCOUNT")
             trancount, = cursor.fetchone()
         if trancount == 0:
             return
-
-        self.adoConn.RollbackTrans()
+        self.transaction_level = self.adoConn.RollbackTrans()
         if not(self.adoConn.Attributes & adXactAbortRetaining):
             #If attributes has adXactAbortRetaining it performs retaining aborts that is,
             #calling RollbackTrans automatically starts a new transaction. Not all providers support this.
             #If not, we will have to start a new transaction by this command:
-            self.adoConn.BeginTrans()
+            self.transaction_level = self.adoConn.BeginTrans()
 
     def cursor(self):
         """Return a new Cursor object using the current connection."""
@@ -332,22 +339,22 @@ class Connection(object):
         return Cursor(self)
 
     def printADOerrors(self):
-        six.print_('ADO Errors (%i):' % self.adoConn.Errors.Count)
+        print('ADO Errors (%i):' % self.adoConn.Errors.Count)
         for e in self.adoConn.Errors:
-            six.print_('Description: %s' % e.Description)
-            six.print_('Error: %s %s ' % (e.Number, adoErrors.get(e.Number, "unknown")))
+            print('Description: %s' % e.Description)
+            print('Error: %s %s ' % (e.Number, adoErrors.get(e.Number, "unknown")))
             if e.Number == ado_error_TIMEOUT:
-                six.print_('Timeout Error: Try using adodbpi.connect(constr,timeout=Nseconds)')
-            six.print_('Source: %s' % e.Source)
-            six.print_('NativeError: %s' % e.NativeError)
-            six.print_('SQL State: %s' % e.SQLState)
-            
+                print('Timeout Error: Try using adodbpi.connect(constr,timeout=Nseconds)')
+            print('Source: %s' % e.Source)
+            print('NativeError: %s' % e.NativeError)
+            print('SQL State: %s' % e.SQLState)
+
     def _suggest_error_class(self):
         """Introspect the current ADO Errors and determine an appropriate error class.
-        
+
         Error.SQLState is a SQL-defined error condition, per the SQL specification:
         http://www.contrib.andrew.cmu.edu/~shadow/sql/sql1992.txt
-        
+
         The 23000 class of errors are integrity errors.
         Error 40002 is a transactional integrity error.
         """
@@ -356,7 +363,7 @@ class Connection(object):
                 state = str(e.SQLState)
                 if state.startswith('23') or state=='40002':
                     return IntegrityError
-            
+
         return DatabaseError
 
     def __del__(self):
@@ -394,15 +401,21 @@ class Cursor(object):
 
     def __iter__(self):
         return iter(self.fetchone, None)
-        
+
     def __enter__(self):
         "Allow database cursors to be used with context managers."
         return self
-        
+
     def __exit__(self, exc_type, exc_val, exc_tb):
         "Allow database cursors to be used with context managers."
         self.close()
 
+    def __del__(self):
+        try:
+            self.close()
+        except:
+        	pass
+
     def _raiseCursorError(self, errorclass, errorvalue):
         eh = self.errorhandler
         if eh is None:
@@ -420,7 +433,7 @@ class Cursor(object):
         self.rowcount = -1
         self.rs = recordset
         desc = list()
-        
+
         for f in self.rs.Fields:
             display_size = None
             if not(self.rs.EOF or self.rs.BOF):
@@ -429,7 +442,7 @@ class Cursor(object):
             null_ok = bool(f.Attributes & adFldMayBeNull)
 
             desc.append( (f.Name, f.Type, display_size, f.DefinedSize, f.Precision, f.NumericScale, null_ok) )
-            
+
         self.description = desc
 
     def close(self):
@@ -445,9 +458,12 @@ class Cursor(object):
         self.messages = []
 
         if self.connection is None:
-            self._raiseCursorError(Error, None)
+            self._raiseCursorError(InterfaceError, None)
             return
 
+        # Inner import to make this module importable on non-Windows platforms.
+        import win32com.client
+
         try:
             self.cmd = win32com.client.Dispatch("ADODB.Command")
             self.cmd.ActiveConnection = self.connection.adoConn
@@ -466,8 +482,8 @@ class Cursor(object):
             self._description_from_recordset(recordset[0])
         except Exception as e:
             _message = ""
-            if hasattr(e, 'args'): _message += str(e.args)
-            #_message += "Command:\n%s\nParameters:\n%s" %  (self.cmd.CommandText, format_parameters(self.cmd.Parameters, True))
+            if hasattr(e, 'args'): _message += str(e.args)+"\n"
+            _message += "Command:\n%s\nParameters:\n%s" %  (self.cmd.CommandText, format_parameters(self.cmd.Parameters, True))
             klass = self.connection._suggest_error_class()
             self._raiseCursorError(klass, _message)
 
@@ -496,7 +512,7 @@ class Cursor(object):
             for i, p in enumerate(tuple(self.cmd.Parameters)[1:]):
                 _configure_parameter(p, parameters[i])
         except:
-            _message = u'Converting Parameter %s: %s, %s\n' %\
+            _message = 'Converting Parameter %s: %s, %s\n' %\
                 (p.Name, ado_type_name(p.Type), repr(parameters[i]))
 
             self._raiseCursorError(DataError, _message)
@@ -525,8 +541,8 @@ class Cursor(object):
             if value is None:
                 parameter_replacements.append('NULL')
                 continue
-                
-            if isinstance(value, basestring) and value == "":
+
+            if isinstance(value, six.string_types) and value == "":
                 parameter_replacements.append("''")
                 continue
 
@@ -535,17 +551,17 @@ class Cursor(object):
             try:
                 p = self.cmd.CreateParameter('p%i' % i, _ado_type(value))
             except KeyError:
-                _message = u'Failed to map python type "%s" to an ADO type' % (value.__class__.__name__,)
+                _message = 'Failed to map python type "%s" to an ADO type' % (value.__class__.__name__,)
                 self._raiseCursorError(DataError, _message)
             except:
-                _message = u'Creating Parameter p%i, %s' % (i, _ado_type(value))
+                _message = 'Creating Parameter p%i, %s' % (i, _ado_type(value))
                 self._raiseCursorError(DataError, _message)
 
             try:
                 _configure_parameter(p, value)
                 self.cmd.Parameters.Append(p)
             except Exception as e:
-                _message = u'Converting Parameter %s: %s, %s\n' %\
+                _message = 'Converting Parameter %s: %s, %s\n' %\
                     (p.Name, ado_type_name(p.Type), repr(value))
 
                 self._raiseCursorError(DataError, _message)
@@ -632,13 +648,13 @@ class Cursor(object):
         """
         self.messages = list()
         if self.connection is None or self.rs is None:
-            #self._raiseCursorError(Error, None)
+            self._raiseCursorError(Error, None)
             return None
 
         recordset = self.rs.NextRecordset()[0]
         if recordset is None:
             return None
-            
+
         self._description_from_recordset(recordset)
         return True
 
@@ -649,7 +665,7 @@ class Cursor(object):
 Date = datetime.date
 Time = datetime.time
 Timestamp = datetime.datetime
-Binary = buffer
+Binary = six.memoryview
 
 def DateFromTicks(ticks):
     """Construct an object holding a date value from the given # of ticks."""
@@ -703,14 +719,17 @@ def _convertNumberWithCulture(variant, f):
         except (ValueError,TypeError): pass
 
 def _cvtComDate(comDate):
-    date_as_float = float(comDate)
-    day_count = int(date_as_float)
-    fraction_of_day = abs(date_as_float - day_count)
+    if isinstance(comDate, datetime.datetime):
+        dt = comDate
+    else:
+        date_as_float = float(comDate)
+        day_count = int(date_as_float)
+        fraction_of_day = abs(date_as_float - day_count)
 
-    dt = (datetime.datetime.fromordinal(day_count + _ordinal_1899_12_31) +
-        datetime.timedelta(milliseconds=fraction_of_day * _milliseconds_per_day))
-    
-    if timezone and getattr(settings, 'USE_TZ', False):
+        dt = (datetime.datetime.fromordinal(day_count + _ordinal_1899_12_31) +
+            datetime.timedelta(milliseconds=fraction_of_day * _milliseconds_per_day))
+
+    if getattr(settings, 'USE_TZ', False):
         dt = dt.replace(tzinfo=timezone.utc)
     return dt
 
@@ -720,26 +739,31 @@ _variantConversions = MultiMap(
         adoExactNumericTypes: _cvtDecimal,
         adoApproximateNumericTypes: _cvtFloat,
         (adBoolean,): bool,
-        adoLongTypes+adoRowIdTypes : long,
+        adoLongTypes+adoRowIdTypes : int if six.PY3 else long,
         adoIntegerTypes: int,
-        adoBinaryTypes: buffer,
-    }, 
+        adoBinaryTypes: six.memoryview,
+    },
     lambda x: x)
 
 # Mapping Python data types to ADO type codes
 def _ado_type(data):
-    if isinstance(data, basestring):
+    if isinstance(data, six.string_types):
         return adBSTR
     return _map_to_adotype[type(data)]
 
 _map_to_adotype = {
-    buffer: adBinary,
+    six.memoryview: adBinary,
     float: adDouble,
-    int: adInteger,
-    long: adBigInt,
+    int: adInteger if six.PY2 else adBigInt,
     bool: adBoolean,
     decimal.Decimal: adDecimal,
     datetime.date: adDate,
     datetime.datetime: adDate,
-    datetime.time: adDate, 
+    datetime.time: adDate,
 }
+
+if six.PY3:
+    _map_to_adotype[bytes] = adBinary
+
+if six.PY2:
+    _map_to_adotype[long] = adBigInt
