@@ -1,4 +1,4 @@
-from __future__ import absolute_import
+from __future__ import absolute_import, unicode_literals
 
 try:
     from itertools import zip_longest
@@ -6,8 +6,11 @@ except ImportError:
     from itertools import izip_longest as zip_longest
 
 import django
+from django.db.utils import DatabaseError
+from django.db.transaction import TransactionManagementError
 from django.db.models.sql import compiler
 import re
+import six
 from itertools import chain, repeat
 
 # query_class returns the base class to use for Django queries.
@@ -21,6 +24,8 @@ from itertools import chain, repeat
 _re_order_limit_offset = re.compile(
     r'(?:ORDER BY\s+(.+?))?\s*(?:LIMIT\s+(\d+))?\s*(?:OFFSET\s+(\d+))?$')
 
+_re_find_order_direction = re.compile(r'\s+(asc|desc)\s*$', re.IGNORECASE)
+
 # Pattern to find the quoted column name at the end of a field specification
 _re_pat_col = re.compile(r"\[([^\[]+)\]$")
 
@@ -65,26 +70,8 @@ def _remove_order_limit_offset(sql):
 
 
 class SQLCompiler(compiler.SQLCompiler):
-    def __pad_fields_with_aggregates(self, fields):
-        """
-        Fix for Django ticket #21126 backported to Django 1.5-1.5.4
-        """
-        if django.VERSION[:2] == (1, 5) and django.VERSION[2] <= 5:
-            if not bool(self.query.aggregate_select):
-                return fields
-            aggregate_start = len(self.query.extra_select) + len(self.query.select)
-            aggregate_end = aggregate_start + len(self.query.aggregate_select)
-
-            # pad None in to fields for aggregates
-            return fields[:aggregate_start] + [
-                None for x in range(0, aggregate_end - aggregate_start)
-            ] + fields[aggregate_start:]
-        return fields
-
     def resolve_columns(self, row, fields=()):
-        fields = self.__pad_fields_with_aggregates(fields)
-
-        # If the results are sliced, the resultset will have an initial 
+        # If the results are sliced, the resultset will have an initial
         # "row number" column. Remove this column before the ORM sees it.
         if getattr(self, '_using_row_number', False):
             row = row[1:]
@@ -110,22 +97,110 @@ class SQLCompiler(compiler.SQLCompiler):
         E.g. AVG([1, 2]) needs to yield 1.5, not 1
         """
         for alias, aggregate in self.query.aggregate_select.items():
-            if aggregate.sql_function == 'AVG' and self.connection.cast_avg_to_float:
+            sql_function = getattr(aggregate, 'sql_function', None)
+            if not sql_function:
+                continue
+            if sql_function == 'AVG' and self.connection.cast_avg_to_float:
                 # Embed the CAST in the template on this query to
                 # maintain multi-db support.
                 self.query.aggregate_select[alias].sql_template = \
                     '%(function)s(CAST(%(field)s AS FLOAT))'
             # translate StdDev function names
-            elif aggregate.sql_function == 'STDDEV_SAMP':
+            elif sql_function == 'STDDEV_SAMP':
                 self.query.aggregate_select[alias].sql_function = 'STDEV'
-            elif aggregate.sql_function == 'STDDEV_POP':
+            elif sql_function == 'STDDEV_POP':
                 self.query.aggregate_select[alias].sql_function = 'STDEVP'
             # translate Variance function names
-            elif aggregate.sql_function == 'VAR_SAMP':
+            elif sql_function == 'VAR_SAMP':
                 self.query.aggregate_select[alias].sql_function = 'VAR'
-            elif aggregate.sql_function == 'VAR_POP':
+            elif sql_function == 'VAR_POP':
                 self.query.aggregate_select[alias].sql_function = 'VARP'
 
+    def _as_sql(self, with_col_aliases=False):
+        """
+        Creates the SQL for this query. Returns the SQL string and list of
+        parameters.
+
+        """
+        self.pre_sql_setup()
+        # After executing the query, we must get rid of any joins the query
+        # setup created. So, take note of alias counts before the query ran.
+        # However we do not want to get rid of stuff done in pre_sql_setup(),
+        # as the pre_sql_setup will modify query state in a way that forbids
+        # another run of it.
+        self.refcounts_before = self.query.alias_refcount.copy()
+        out_cols, s_params = self.get_columns(with_col_aliases)
+        ordering, o_params, ordering_group_by = self.get_ordering()
+
+        distinct_fields = self.get_distinct()
+
+        # This must come after 'select', 'ordering' and 'distinct' -- see
+        # docstring of get_from_clause() for details.
+        from_, f_params = self.get_from_clause()
+
+        qn = self.quote_name_unless_alias
+
+        where, w_params = self.query.where.as_sql(qn=qn, connection=self.connection)
+        having, h_params = self.query.having.as_sql(qn=qn, connection=self.connection)
+        if hasattr(self.query.having, 'get_group_by_cols'):  # Django 1.7
+            having_group_by = self.query.having.get_group_by_cols()
+        else:
+            having_group_by = self.query.having.get_cols()
+        params = []
+        for val in six.itervalues(self.query.extra_select):
+            params.extend(val[1])
+
+        result = ['SELECT']
+
+        if self.query.distinct:
+            result.append(self.connection.ops.distinct_sql(distinct_fields))
+        params.extend(o_params)
+        result.append(', '.join(out_cols + self.ordering_aliases))
+        params.extend(s_params)
+        params.extend(self.ordering_params)
+
+        result.append('FROM')
+        result.extend(from_)
+        params.extend(f_params)
+
+        if self.query.select_for_update and self.connection.features.has_select_for_update:
+            if self.connection.get_autocommit():
+                raise TransactionManagementError("mssql_select_for_update cannot be used outside of a transaction.")
+
+            # If we've been asked for a NOWAIT query but the backend does not support it,
+            # raise a DatabaseError otherwise we could get an unexpected deadlock.
+            nowait = self.query.select_for_update_nowait
+            if nowait and not self.connection.features.has_select_for_update_nowait:
+                raise DatabaseError('NOWAIT is not supported on this database backend.')
+            result.append(self.connection.ops.for_update_sql(nowait=nowait))
+
+        if where:
+            result.append('WHERE %s' % where)
+            params.extend(w_params)
+
+        grouping, gb_params = self.get_grouping(having_group_by, ordering_group_by)
+        if grouping:
+            if distinct_fields:
+                raise NotImplementedError(
+                    "annotate() + distinct(fields) not implemented.")
+            if not ordering:
+                ordering = self.connection.ops.force_no_ordering()
+            result.append('GROUP BY %s' % ', '.join(grouping))
+            params.extend(gb_params)
+
+        if having:
+            result.append('HAVING %s' % having)
+            params.extend(h_params)
+
+        if ordering:
+            result.append('ORDER BY %s' % ', '.join(ordering))
+
+
+        # Finally do cleanup - get rid of the joins we created above.
+        self.query.reset_refcounts(self.refcounts_before)
+
+        return ' '.join(result), tuple(params)
+
     def as_sql(self, with_limits=True, with_col_aliases=False):
         # Django #12192 - Don't execute any DB query when QS slicing results in limit 0
         if with_limits and self.query.low_mark == self.query.high_mark:
@@ -143,13 +218,13 @@ class SQLCompiler(compiler.SQLCompiler):
             # unless TOP or FOR XML is also specified.
             try:
                 setattr(self.query, '_mssql_ordering_not_allowed', with_col_aliases)
-                result = super(SQLCompiler, self).as_sql(with_limits, with_col_aliases)
+                result = self._as_sql(with_col_aliases)
             finally:
                 # remove in case query is every reused
                 delattr(self.query, '_mssql_ordering_not_allowed')
             return result
 
-        raw_sql, fields = super(SQLCompiler, self).as_sql(False, with_col_aliases)
+        raw_sql, fields = self._as_sql(with_col_aliases)
         
         # Check for high mark only and replace with "TOP"
         if self.query.high_mark is not None and not self.query.low_mark:
@@ -163,36 +238,17 @@ class SQLCompiler(compiler.SQLCompiler):
         # Else we have limits; rewrite the query using ROW_NUMBER()
         self._using_row_number = True
 
-        order, limit_ignore, offset_ignore = _get_order_limit_offset(raw_sql)
-        
-        qn = self.connection.ops.quote_name
-        
-        inner_table_name = qn('AAAA')
-
-        # Using ROW_NUMBER requires an ordering
-        if order is None:
-            meta = self.query.get_meta()                
-            column = meta.pk.db_column or meta.pk.get_attname()
-            order = '{0}.{1} ASC'.format(inner_table_name, qn(column))
-        else:
-            # remap order for injected subselect
-            new_order = []
-            for x in order.split(','):
-                if x.find('.') != -1:
-                    tbl, col = x.rsplit('.', 1)
-                else:
-                    col = x
-                new_order.append('{0}.{1}'.format(inner_table_name, col))
-            order = ', '.join(new_order)
-        
-        where_row_num = '{0} < _row_num'.format(self.query.low_mark)
-        if self.query.high_mark:
-            where_row_num += ' and _row_num <= {0}'.format(self.query.high_mark)
-            
         # Lop off ORDER... and the initial "SELECT"
         inner_select = _remove_order_limit_offset(raw_sql)
         outer_fields, inner_select = self._alias_columns(inner_select)
 
+        order = _get_order_limit_offset(raw_sql)[0]
+
+        qn = self.connection.ops.quote_name
+        inner_table_name = qn('AAAA')
+
+        outer_fields, inner_select, order = self._fix_slicing_order(outer_fields, inner_select, order, inner_table_name)
+
         # map a copy of outer_fields for injected subselect
         f = []
         for x in outer_fields.split(','):
@@ -211,7 +267,11 @@ class SQLCompiler(compiler.SQLCompiler):
             inner=inner_select,
             inner_as=inner_table_name,
         )
-        
+
+        where_row_num = '{0} < _row_num'.format(self.query.low_mark)
+        if self.query.high_mark:
+            where_row_num += ' and _row_num <= {0}'.format(self.query.high_mark)
+
         sql = "SELECT _row_num, {outer} FROM ( SELECT ROW_NUMBER() OVER ( ORDER BY {order}) as _row_num, {inner}) as QQQ where {where}".format(
             outer=outer_fields,
             order=order,
@@ -221,6 +281,54 @@ class SQLCompiler(compiler.SQLCompiler):
         
         return sql, fields
 
+    def _fix_slicing_order(self, outer_fields, inner_select, order, inner_table_name):
+        """
+        Apply any necessary fixes to the outer_fields, inner_select, and order
+        strings due to slicing.
+        """
+        # Using ROW_NUMBER requires an ordering
+        if order is None:
+            meta = self.query.get_meta()
+            column = meta.pk.db_column or meta.pk.get_attname()
+            order = '{0}.{1} ASC'.format(
+                inner_table_name,
+                self.connection.ops.quote_name(column),
+            )
+        else:
+            alias_id = 0
+            # remap order for injected subselect
+            new_order = []
+            for x in order.split(','):
+                # find the ordering direction
+                m = _re_find_order_direction.search(x)
+                if m:
+                    direction = m.groups()[0]
+                else:
+                    direction = 'ASC'
+                # remove the ordering direction
+                x = _re_find_order_direction.sub('', x)
+                # remove any namespacing or table name from the column name
+                col = x.rsplit('.', 1)[-1]
+                # Is the ordering column missing from the inner select?
+                # 'inner_select' contains the full query without the leading 'SELECT '.
+                # It's possible that this can get a false hit if the ordering
+                # column is used in the WHERE while not being in the SELECT. It's
+                # not worth the complexity to properly handle that edge case.
+                if x not in inner_select:
+                    # Ordering requires the column to be selected by the inner select
+                    alias_id += 1
+                    # alias column name
+                    col = '[{0}___o{1}]'.format(
+                        col.strip('[]'),
+                        alias_id,
+                    )
+                    # add alias to inner_select
+                    inner_select = '({0}) AS {1}, {2}'.format(x, col, inner_select)
+                new_order.append('{0}.{1} {2}'.format(inner_table_name, col, direction))
+
+            order = ', '.join(new_order)
+        return outer_fields, inner_select, order
+
     def _alias_columns(self, sql):
         """Return tuple of SELECT and FROM clauses, aliasing duplicate column names."""
         qn = self.connection.ops.quote_name
@@ -250,21 +358,8 @@ class SQLCompiler(compiler.SQLCompiler):
     
         def _replace_sub(col):
             """Replace all placeholders with expanded values"""
-            while True:
-                m = _re_col_placeholder.search(col)
-                if m:
-                    try:
-                        key = '_placeholder_{0}'.format(
-                            int(m.group(1))
-                        )
-                        col = col.format(**{
-                            key : parens[key]
-                        })
-                    except:
-                        # not a substituted value
-                        break
-                else:
-                    break
+            while _re_col_placeholder.search(col):
+                col = col.format(**parens)
             return col
     
         temp_sql = ''.join(paren_buf)
@@ -280,16 +375,11 @@ class SQLCompiler(compiler.SQLCompiler):
                 if col_key in names_seen:
                     alias = qn('{0}___{1}'.format(col_name, names_seen.count(col_key)))
                     outer.append(alias)
-            
-                    col = _replace_sub(col)
-            
-                    inner.append('{0} as {1}'.format(col, alias))
+                    inner.append('{0} as {1}'.format(_replace_sub(col), alias))
                 else:
-                    replaced = _replace_sub(col)
-                            
                     outer.append(qn(col_name))
-                    inner.append(replaced)
-    
+                    inner.append(_replace_sub(col))
+
                 names_seen.append(col_key)
             else:
                 raise Exception('Unable to find a column name when parsing SQL: {0}'.format(col))
